"""CLI wrapper for cryptographic functions of the :mod:`restful` module.

To be called via :func:`subprocess.run()` as a workaround for
:class:`ImportError`s related to PyO3's current lack of sub-interpreter
support.

Note:
    Since this module is installed as part of the ``ceph`` package,
    it should be called like so::

        python3 -m ceph._crypto_wrap create_self_signed_cert
"""

import argparse
import sys
import json

from argparse import Namespace
from typing import Any
from uuid import uuid4

from OpenSSL import crypto


def _respond(data: dict[str, Any]) -> None:
    json.dump(data, sys.stdout)
    sys.stdout.flush()


def create_self_signed_cert(args: Namespace) -> None:
    cert_key_pair = _create_self_signed_cert()
    _respond(cert_key_pair)


def _create_self_signed_cert() -> dict[str, str]:
    # create a key pair
    pubkey = crypto.PKey()
    pubkey.generate_key(crypto.TYPE_RSA, 2048)

    # create a self-signed cert
    cert = crypto.X509()
    cert.get_subject().O = "IT"
    cert.get_subject().CN = "ceph-restful"
    cert.set_serial_number(int(uuid4()))
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(10 * 365 * 24 * 60 * 60)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(pubkey)
    cert.sign(pubkey, "sha512")

    return {
        "cert": crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode(),
        "key": crypto.dump_privatekey(crypto.FILETYPE_PEM, pubkey).decode(),
    }


def main() -> None:
    parser = argparse.ArgumentParser(prog="_crypto_wrap.py")
    subparsers = parser.add_subparsers(required=True)

    parser_cssc = subparsers.add_parser("create_self_signed_cert")
    parser_cssc.set_defaults(func=create_self_signed_cert)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()
